@@ -2,7 +2,7 @@
 # python2.5 /usr/lib/hildon-desktop/recaller.py
 
 APP_TITLE = "Recaller"
-APP_VERSION = "2.0"
+APP_VERSION = "2.1"
 
 import gobject
 import gtk
@@ -15,9 +15,16 @@ import osso
 import string
 import time 
 import gconf
+import urllib
+import sys
+
+import ctypes
+import gettext
+import evolution
 
 import dbus
 
+gobject.threads_init ()
 import pygst
 pygst.require("0.10")
 import gst
@@ -47,13 +54,16 @@ class RecallerHomePlugin(hildondesktop.HomePluginItem):
     self.defaultAUTOREC = False
     self.defaultBANNERS = True
 
-    self.SHOWBANNER = True
+    #self.SHOWBANNER = False
     
     self.isStopped = True 
     self.update_info()
     self.transparency = 0.0
     self.timer = None
 
+	#TODO get that working...argh
+    #self.cb = ContactHandler()
+    
     #Load the settings using GConf
     conf_client = gconf.client_get_default()
     conf_client.add_dir(GCONFPATH, gconf.CLIENT_PRELOAD_NONE)
@@ -85,17 +95,14 @@ class RecallerHomePlugin(hildondesktop.HomePluginItem):
       
     self.isStopped = True 
     self.transparency = 0.0 
-
    
     #The usual way IF NOT a hildon-home widget. Will work when in an app
     #self.bus = dbus.SystemBus()
     #The suggestion from Kimmo Hämäläinen, doesn't help, works the same as plain "self.bus = dbus.SystemBus()" above
     #dbus_loop = DBusGMainLoop()
-    #self.bus = dbus.SystemBus(mainloop=dbus_loop)
-    
-    #Another variant, refuses to connect (PUT EXACT MESSAGE HERE)
+    #self.bus = dbus.SystemBus(mainloop=dbus_loop)   
+    #This works :-)
     self.bus = dbus.bus.BusConnection("unix:path=/var/run/dbus/system_bus_socket")
-    #RuntimeError: To make asynchronous calls, receive signals or export objects, D-Bus connections must be attached to a main loop by passing mainloop=... to the constructor or calling dbus.set_default_main_loop(...)
 
     #---Incoming/outgoing callbacks---------------------------
 
@@ -107,7 +114,7 @@ class RecallerHomePlugin(hildondesktop.HomePluginItem):
                             bus_name="com.nokia.csd",
                             path="/com/nokia/csd/call",
                             )
-    
+                                
     #[SIGNAL] com.nokia.csd.Call.Instance.Terminated  /com/nokia/csd/call/1  :1.17
     #()
     self.bus.add_signal_receiver(self.onHook,
@@ -116,7 +123,8 @@ class RecallerHomePlugin(hildondesktop.HomePluginItem):
                             bus_name="com.nokia.csd",
                             path="/com/nokia/csd/call/1",
                             )
-    
+                            
+                        
     #---Incoming callbacks------------------------------------
     
     #[SIGNAL] com.nokia.csd.Call.Coming  /com/nokia/csd/call  :1.17
@@ -137,7 +145,7 @@ class RecallerHomePlugin(hildondesktop.HomePluginItem):
                             bus_name="com.nokia.csd",
                             path="/com/nokia/csd/call",
                             )
-                            
+    
     root = self.bus.get_object('org.bluez', '/')
     manager = dbus.Interface(root, 'org.bluez.Manager')
     defaultAdapter = manager.DefaultAdapter()    
@@ -220,31 +228,46 @@ class RecallerHomePlugin(hildondesktop.HomePluginItem):
     colormap = screen.get_rgba_colormap()
     self.set_colormap(colormap)
     hildondesktop.HomePluginItem.do_realize(self) 
-   
   
   def on_button_clicked(self, widget, data=None):
-      # do not listen if in middle of state change
+      #The recording is currently running, we must stop  it
       if not self.isStopped:
         success, state, pending = self.recorder.get_state(1)
-        if not pending:
-            if state == gst.STATE_PLAYING:
-              self.stop()              
+        # do not listen if in middle of state change
+        #if not pending:
+        if state == gst.STATE_PLAYING:
+          self.stop() 
+        elif state == gst.STATE_NULL: #Avoid hangs if something went wrong
+          #self.stop()  
+          self.draw_e_off()
+          self.isStopped=True
+          #self.recorder.send_event(gst.event_new_eos())
+      #No recording going on yet, let's start a manual recording
       else:
-          self.caller = None
-          self.start_PhoneRecording()
-          self.isStopped = False
+        self.caller = None
+        self.start_PhoneRecording()
           
   def start_PhoneRecording(self):
-      self.draw_e_on()
       self.startTime =  time.strftime("%Y%m%d_%H%M%S")
       self.startTimeToDisplay =  time.strftime("%H:%M:%S")
       if self.caller == None:
         self.fileName = os.path.join(self.folder,"Rec_%s.%s" % (self.startTime,self.fileformat.lower()))
       else:
-        self.fileName = os.path.join(self.folder,"Rec_%s_%s.%s" % (self.caller,self.startTime,self.fileformat.lower()))     
+        self.fileName = os.path.join(self.folder,"Rec_%s_%s.%s" % (self.caller,self.startTime,self.fileformat.lower()))  
+      self.fileName = os.path.join(self.folder,"CurrentlyRecording.%s" % self.fileformat.lower())
+      print "Self.source="+self.source 
+      print "self.fileName="+self.fileName 
       print "Launcher config is:"
       print self.get_launcher()
+      #self.gsterr = ""
       self.recorder = gst.parse_launch ( self.get_launcher())
+      #print "parse_error:"+str(self.gsterr)
+      if self.recorder == None:
+        print "Launch failure!"
+        return
+      # else:
+        # print "launch state"
+        # print self.recorder.get_state(1)
       bus = self.recorder.get_bus()
       # need to connect to sync message handler so we get the sink to be
       # embedded at the right time and not have a temporary new window
@@ -253,24 +276,32 @@ class RecallerHomePlugin(hildondesktop.HomePluginItem):
       bus.connect("sync-message::element", self.on_sync_message)
       bus.connect("message", self.on_message)
       self.recorder.set_state(gst.STATE_PLAYING)
+      #success, state, pending = self.recorder.get_state(1)
+      #if state != gst.STATE_PLAYING:
+      #  self.log("ERROR: Failed to start recording in gstreamer!")
+      #  return
+      self.draw_e_on()
+      self.isStopped = False
       if self.reminders:
         self.timer = gobject.timeout_add_seconds(30, self.recordingreminder_cb)      
-
+      # print "phonerecording!"
       
   def stop(self):
       if not self.timer is None:     
         gobject.source_remove(self.timer)
-      self.draw_e_off()
       self.recorder.send_event(gst.event_new_eos())
+      
+  def cleanup(self):
+      self.isStopped=True
+      self.draw_e_off()
       if self.caller == None:
         finalName = os.path.join(self.folder,"Rec_%s-%s.%s" % (self.startTime, time.strftime("%H%M%S"),self.fileformat.lower()))
       else:
         finalName = os.path.join(self.folder,"Rec_%s_%s-%s.%s" % (self.caller,self.startTime,time.strftime("%H%M%S"),self.fileformat.lower()))       
       self.log(finalName) 
       os.rename(self.fileName, finalName)
-      self.isStopped=True
       self.log("recaller: " + time.strftime("%H:%M:%S") + " - Recording stopped.")     
-
+      
   def on_sync_message(self, bus, message):
       if message.structure is None:
           return
@@ -287,19 +318,24 @@ class RecallerHomePlugin(hildondesktop.HomePluginItem):
       if message.type == gst.MESSAGE_ERROR:
         self.recorder.set_state(gst.STATE_NULL)
         err, debug = message.parse_error()
+        print "GST ERROR"
+        print str(err)
+        print str(debug)
+        self.cleanup()
       if message.type == gst.MESSAGE_EOS:
         self.recorder.set_state(gst.STATE_NULL)
-
+        self.cleanup()
+      
   def get_launcher(self):
       enc = self.get_encoder()
       if self.source == "Phone":
-        self.launcher = "adder name=theAdder ! %s ! filesink location=%s. pulsesrc device=sink.hw0.monitor ! queue name=q1 ! theAdder. pulsesrc device=source.hw0 ! queue name=q2 ! theAdder."  % (enc,self.fileName)
+        self.launcher = "adder name=theAdder ! %s ! filesink location=\"%s\" . pulsesrc device=sink.hw0.monitor ! queue name=q1 ! theAdder. pulsesrc device=source.hw0 ! audioamplify amplification=30.0 ! queue name=q2 ! theAdder."  % (enc,self.fileName)
       elif self.source == "Bluetooth":
-        self.launcher = "adder name=theAdder ! %s ! filesink location=%s. pulsesrc device=sink.hw1.monitor ! queue name=q1 ! theAdder. pulsesrc device=source.hw1 ! queue name=q2 ! theAdder."  % (enc,self.fileName)
+        self.launcher = "adder name=theAdder ! %s ! filesink location=\"%s\" . pulsesrc device=sink.hw1.monitor ! audioamplify amplification=10.0 ! queue name=q1 ! theAdder. pulsesrc device=source.hw1 ! audioamplify amplification=5.0 ! queue name=q2 ! theAdder."  % (enc,self.fileName)
       elif self.source == "Dictaphone":
-        self.launcher = "pulsesrc device=source.hw0 ! audioconvert ! queue ! %s ! filesink location=%s" % (enc,self.fileName)
+        self.launcher = "pulsesrc device=source.hw0 ! audioconvert ! audioamplify amplification=30.0 ! queue ! %s ! filesink location=\"%s\"" % (enc,self.fileName)
       elif self.source == "System":
-        self.launcher = "pulsesrc device=sink.hw0.monitor ! audioconvert ! queue ! %s ! filesink location=%s" % (enc,self.fileName)
+        self.launcher = "pulsesrc device=sink.hw0.monitor ! audioconvert ! queue ! %s ! filesink location=\"%s\"" % (enc,self.fileName)
       return self.launcher
 
   def get_encoder(self):
@@ -322,8 +358,8 @@ class RecallerHomePlugin(hildondesktop.HomePluginItem):
       if rep==gtk.RESPONSE_OK:
           print "new selection=%s"%self.folder
           self.btnOpen.set_text("Save Folder", self.folder[0:16])
-      else:
-          print "no change"
+      # else:
+          # print "no change"
       
   def format_changed(self, widget, data):
       self.fileformat = self.selectFormat.get_current_text().lower()
@@ -342,6 +378,17 @@ class RecallerHomePlugin(hildondesktop.HomePluginItem):
 
   def callInfoFrom(self,sender, message):
     if sender == "/com/nokia/csd/call/1":
+      if message == None:
+        message = "Unknown"
+      message = message.replace ( '*', '+' )
+      # print message
+      
+      # senderuid = self.cb.get_uid_from_number(message)
+      # if senderuid != None:
+        # sender = self.cb.get_displayname_from_uid(senderuid)
+      # if sender != None:
+        # sndr = sender
+      # print sndr
       self.caller = str(message)
       self.log("Phone call from %s" % self.caller)
 
@@ -355,26 +402,64 @@ class RecallerHomePlugin(hildondesktop.HomePluginItem):
       self.log("Picked up the call...")
       if self.autorec:
         self.currentsource = self.source  #Store original audio source
-        if self.source == "Dictaphone" or self.source == "System":
-          if self.hasHandsfreeConnected():
-            self.source == "Bluetooth"
-          else:
-            self.source == "Phone"
+        #Switch to handsfree if connected
+        self.hasHandsfree = self.hasHandsfreeConnected()
+        if self.hasHandsfree == True:
+          self.source = "Bluetooth" 
+          self.log("Self.source set to Bluetooth")          
+        else:
+          self.source = "Phone"
+          self.log("Self.source set to Phone")
+        #elif self.source == "Dictaphone" or self.source == "System":
+          # sourceSwitchingNeeded = True
+          # self.log("hasHandsfree="+str(hasHandsfree))
+          # if hasHandsfree:
+            # self.source == "Bluetooth"
+            # self.log("Self.source set to Bluetooth")
+          # else:
+          #self.source == "Phone"
+    
+        #Special case: a recording is already running
+        if self.isStopped == False:
+          if self.currentsource == self.source: #Just abort and continue recording
+            self.log("Recording already running, no source change, just continue")
+            return
+          else:                         #Stop the current recording and restart new one
+            success, state, pending = self.recorder.get_state(1)
+            if not pending:
+              if state == gst.STATE_PLAYING:
+                self.stop()
+                self.log("Recording already running,source changed, stop and restart just after that")
+            
         self.start_PhoneRecording()
         self.isStopped = False
-        
+            
   def onHook(self):
     #print sender, "said:", message
     #if sender == "/com/nokia/csd/call/1":
     self.log("Call terminated...")
-    if not self.isStopped:
-      #self.caller = None       #Bad idea, makes renaming fail...
+    if not self.isStopped:  #Recording still running, let's stop
+      try:
+        if(self.source == "Bluetooth"): #Emergency stop or we will crash when the BT source/sink disappear moments later...
+          if not self.recorder is None:
+            self.recorder.set_state(gst.STATE_NULL)
+          self.source = self.currentsource
+          if not self.timer is None:     
+            gobject.source_remove(self.timer)
+          self.cleanup()
+          self.log("Did emergency full stop due to BT headset being used :-)")
+          return
+      except:
+        print(sys.exc_info()[0])
+      #For all others, clean up a bit more properly
       success, state, pending = self.recorder.get_state(1)
       if not pending:
-          if state == gst.STATE_PLAYING:
-            self.stop()
-            self.source = self.currentsource  #Restore original audio source        
-            
+        if state == gst.STATE_PLAYING:
+          self.source = self.currentsource  #Restore original audio source  
+          self.stop()
+      else:
+         self.log("onHook.pending, can't stop?")
+      
   def hasHandsfreeConnected(self):
     #If BT is disabled, we can already abort
     props = self.BTadapter.GetProperties()
@@ -395,12 +480,12 @@ class RecallerHomePlugin(hildondesktop.HomePluginItem):
       isConnected = devProps['Connected']
       #Check if a connected device offers a handsfree service
       if isConnected:
-        print devName
+        #print devName
         UUIDs = devProps['UUIDs']
         for uuid in UUIDs:
           id = uuid[4:8]
           if id == "111e":
-            print " is a connected handsfree device !"
+            #print " is a connected handsfree device !"
             return True
     return False  
       
@@ -465,7 +550,6 @@ class RecallerHomePlugin(hildondesktop.HomePluginItem):
 
       #Btn to select the default save folder
       self.btnOpen = hildon.Button(gtk.HILDON_SIZE_FULLSCREEN_WIDTH | gtk.HILDON_SIZE_FINGER_HEIGHT,hildon.BUTTON_ARRANGEMENT_VERTICAL)
-      #self.btnOpen.set_text("Save Folder", self.folder[0:16])
       self.btnOpen.set_text("Save Folder", self.folder)
       self.btnOpen.connect_after("clicked", self.choose_folder)
 
@@ -514,7 +598,7 @@ class RecallerHomePlugin(hildondesktop.HomePluginItem):
         self.btnAutoRec.set_active(True)
 
       self.btnReminders = hildon.CheckButton(gtk.HILDON_SIZE_HALFSCREEN_WIDTH | gtk.HILDON_SIZE_FINGER_HEIGHT)
-      self.btnReminders.set_label("Show reminders")
+      self.btnReminders.set_label("Show notifications")
       if self.reminders:
         self.btnReminders.set_active(True)
       
@@ -558,14 +642,18 @@ class RecallerHomePlugin(hildondesktop.HomePluginItem):
       dialog.destroy()
 
   def choose_folder(self,widget):
-      selector=hildon.FileChooserDialog(self, gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER, hildon.FileSystemModel())
-      selector.set_current_folder_uri(self.folder)
+      model = hildon.FileSystemModel()
+      selector=hildon.FileChooserDialog(self, gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER,model)
+      selector.set_current_folder_uri("file://"+self.folder)
+      #selector.set_current_folder("/home/user/MyDocs")
+      #selector.set_show_hidden(True)
       rep=selector.run()
       selector.hide()
       self.folder=selector.get_filename()
       if rep==gtk.RESPONSE_OK:
           print "new selection=%s"%self.folder
-          self.btnOpen.set_text("Save Folder", self.folder[0:16])
+          #self.btnOpen.set_text("Save Folder", self.folder[0:16])
+          self.btnOpen.set_text("Save Folder", self.folder)
       else:
           print "no change"
      
@@ -601,11 +689,112 @@ License along with this program.  If not, see
   #Log and/or display banner     
   def log(self, message, *args, **kwargs):
     print time.strftime('%H:%M:%S'), '- '+APP_TITLE+':', message % args
-    if self.SHOWBANNER:
-        banner = hildon.hildon_banner_show_information(self, '', APP_TITLE+': ' + message%args)
-        banner.set_timeout(2000)
+    if self.reminders:
+      banner = hildon.hildon_banner_show_information(self, '', APP_TITLE+': ' + message%args)
+      banner.set_timeout(2000)
+    
+
+class ContactHandler:
+
+	def __init__(self):
+		""" Load the evolution addressbook and prepare ctypes for ossoabook. """
+		self.ab = evolution.ebook.open_addressbook("default")
+		self.osso_ctx = osso.Context("recaller", "0.1")
+		self.osso_abook = ctypes.CDLL('libosso-abook-1.0.so.0')
+		empty = ""
+		argv_type = ctypes.c_char_p * len(empty)
+		argv = argv_type(*empty)
+		argc = ctypes.c_int(len(empty))
+		self.osso_abook.osso_abook_init(ctypes.byref(argc), ctypes.byref(argv), hash(self.osso_ctx))
+		self.ebook = ctypes.CDLL('libebook-1.2.so.5')
+		err = ctypes.c_void_p()
+		self.c_book = self.ebook.e_book_new_default_addressbook(ctypes.byref(err))
+		self.glib = ctypes.CDLL('libglib-2.0.so.0')
+		if not self.ebook.e_book_open(self.c_book, 1, 0):
+		  raise TypeError("Could not open ebook")
+		print "cb init done"
+    
+	def get_displayname_from_number(self, sndr):
+		sndr = str(sndr).replace("/TYPE=PLMN", "")
+		sender = sndr
+		senderuid = self.get_uid_from_number(sndr)
+		if senderuid != None:
+			sender = self.get_displayname_from_uid(senderuid)
+		if sender != None:
+			sndr = sender
+		return sndr
+	
+	def get_displayname_from_uid(self, uid):
+		contact = self.ab.get_contact(str(uid))
+		if contact:
+			contact = contact.get_name()
+		return contact
+
+	def get_uid_from_number(self, number):
+		#Gets the contacts UID from a given phonenumber.
+		#Thanks lizardo for the great example how to implement this.
+		if not self.ebook.e_book_open(self.c_book, 1, 0):
+			raise TypeError("Could not open ebook")
+		
+		# The '1' means fuzzy matching is active
+		c_query = self.osso_abook.osso_abook_query_phone_number(number, 1)
+
+		contacts = ctypes.c_void_p()
+		if not self.ebook.e_book_get_contacts(self.c_book, c_query, ctypes.byref(contacts), 0):
+			raise TypeError("Failed to get query results.")
+
+		for i in self.glist(contacts):
+			E_CONTACT_UID = 1
+			e_contact_get_const = self.ebook.e_contact_get_const
+			e_contact_get_const.restype = ctypes.c_char_p
+			uid = e_contact_get_const(i, E_CONTACT_UID)
+			self.glib.g_list_free(contacts)
+			return uid
+
+		self.glib.g_list_free(contacts)
+		return None
+		
+	def glist(self, addr):
+		""" Implementation of GList. """
+		class _GList(ctypes.Structure):
+			_fields_ = [('data', ctypes.c_void_p),
+				    ('next', ctypes.c_void_p)]
+		l = addr
+		while l:
+			if type(l) == int:
+				l = _GList.from_address(l)
+			else:
+				l = _GList.from_address(l.value)
+			yield l.data
+			l = l.next
+		
+# ctypes wrapper for pygobject_new(), based on code snippet from
+# http://faq.pygtk.org/index.py?req=show&file=faq23.041.htp
+class _PyGObject_Functions(ctypes.Structure):
+	_fields_ = [
+		('register_class',
+		    ctypes.PYFUNCTYPE(ctypes.c_void_p, ctypes.c_char_p,
+		    ctypes.c_int, ctypes.py_object, ctypes.py_object)),
+		('register_wrapper',
+		    ctypes.PYFUNCTYPE(ctypes.c_void_p, ctypes.py_object)),
+		('register_sinkfunc',
+		    ctypes.PYFUNCTYPE(ctypes.py_object, ctypes.c_void_p)),
+		('lookupclass',
+		    ctypes.PYFUNCTYPE(ctypes.py_object, ctypes.c_int)),
+		('newgobj',
+		    ctypes.PYFUNCTYPE(ctypes.py_object, ctypes.c_void_p)),
+	]
+
+class PyGObjectCPAI(object):
+	def __init__(self):
+		import gobject
+		py_obj = ctypes.py_object(gobject._PyGObject_API)
+		addr = ctypes.pythonapi.PyCObject_AsVoidPtr(py_obj)
+		self._api = _PyGObject_Functions.from_address(addr)
+
+	def pygobject_new(self, addr):
+		return self._api.newgobj(addr)
     
-      
 hd_plugin_type = RecallerHomePlugin
 
 if __name__ == "__main__":
